"use strict";
require("reflect-metadata");
var Vue = require("vue");
var AnnotatedOptions = (function () {
    function AnnotatedOptions() {
        this.props = {};
        this.watch = {};
        this.events = {};
    }
    return AnnotatedOptions;
}());
var AnnotatedOptionsKey = "vue-component-decorator:options";
var DesignTypeKey = "design:type";
var internalHooks = [
    "data",
    "el",
    "init",
    "created",
    "ready",
    "beforeCompile",
    "compiled",
    "beforeDestroy",
    "destroyed",
    "attached",
    "detached",
    "activate"
];
function assign(target, other) {
    Object.keys(other).forEach(function (k) {
        target[k] = other[k];
    });
    return target;
}
function makeComponent(target, option) {
    option = assign({}, option);
    option.name = option.name || target["name"];
    var proto = target.prototype;
    Object.getOwnPropertyNames(proto).filter(function (name) { return name !== "constructor"; }).forEach(function (name) {
        // hooks
        if (internalHooks.indexOf(name) > -1) {
            option[name] = proto[name];
        }
        var descriptor = Object.getOwnPropertyDescriptor(proto, name);
        if (typeof descriptor.value === "function") {
            // methods
            (option.methods || (option.methods = {}))[name] = descriptor.value;
        }
        else if (descriptor.get || descriptor.set) {
            // computed
            (option.computed || (option.computed = {}))[name] = {
                get: descriptor.get,
                set: descriptor.set
            };
        }
    });
    var ann = Reflect.getOwnMetadata(AnnotatedOptionsKey, proto);
    if (ann != null) {
        // props
        option.props = option.props || ann.props;
        // watch
        option.watch = option.watch || ann.watch;
        // events
        option.events = option.events || ann.events;
    }
    // find super
    var superProto = Object.getPrototypeOf(proto);
    var Super = (superProto instanceof Vue
        ? superProto.constructor
        : Vue);
    return Super.extend(option);
}
function getAnnotatedOptions(target) {
    var ann = Reflect.getOwnMetadata(AnnotatedOptionsKey, target);
    if (ann == null) {
        ann = new AnnotatedOptions();
        Reflect.defineMetadata(AnnotatedOptionsKey, ann, target);
    }
    return ann;
}
function defineProp(target, propertyKey, option) {
    // detect design type and set prop validation
    if ("type" in option) {
    }
    else {
        var type = Reflect.getOwnMetadata(DesignTypeKey, target, propertyKey);
        if ([String, Number, Boolean, Function, Array].indexOf(type) > -1) {
            option = assign({ type: type }, option);
        }
    }
    getAnnotatedOptions(target).props[propertyKey] = option;
}
function defineWatch(target, propertyKey, option) {
    var descriptor = Object.getOwnPropertyDescriptor(target, propertyKey);
    if (typeof descriptor.value !== "function") {
        // TODO: show warning
        return;
    }
    getAnnotatedOptions(target).watch[option.name] = {
        handler: descriptor.value,
        deep: option.deep,
        immidiate: option.immediate
    };
}
function defineEvent(target, propertyKey, name) {
    var descriptor = Object.getOwnPropertyDescriptor(target, propertyKey);
    if (typeof descriptor.value !== "function") {
        // TODO: show warning
        return;
    }
    getAnnotatedOptions(target).events[name] = descriptor.value;
}
var prop = function (option) {
    return function (target, propertyKey) { return defineProp(target, propertyKey.toString(), option || {}); };
};
var vueit = {
    component: function (option) {
        return function (target) { return makeComponent(target, option || {}); };
    },
    prop: prop,
    p: prop(),
    pr: prop({ required: true }),
    pd: function (defaultValue) { return prop({ default: defaultValue }); },
    watch: function (option) {
        return function (target, propertyKey) { return defineWatch(target, propertyKey.toString(), typeof option === "string" ? { name: option } : option); };
    },
    on: function (name) {
        return function (target, propertyKey) { return defineEvent(target, propertyKey.toString(), name); };
    }
};
module.exports = vueit;
